package tests

import (
	"fmt"
	"os"
	"path"
	"strings"
	"testing"

	"github.com/davecgh/go-spew/spew"
	"github.com/nspcc-dev/neo-go/pkg/compiler"
	"github.com/nspcc-dev/neo-go/pkg/neotest"
	"github.com/nspcc-dev/neo-go/pkg/neotest/chain"
	"github.com/nspcc-dev/neo-go/pkg/smartcontract/callflag"
	"github.com/nspcc-dev/neo-go/pkg/smartcontract/trigger"
	"github.com/nspcc-dev/neo-go/pkg/vm"
	"github.com/stretchr/testify/require"
)

const ctrPath = "../contract"

// Key for tests
var (
	validKey   = []byte{1, 2, 3, 4, 5}
	invalidKey = []byte{1, 2, 3}
)

func newExecutor(t *testing.T) *neotest.Executor {
	bc, acc := chain.NewSingle(t)
	return neotest.NewExecutor(t, bc, acc, acc)
}

func TestContract_GetNumber(t *testing.T) {
	e := newExecutor(t)
	ctrGetNum := neotest.CompileFile(t, e.CommitteeHash, ctrPath, path.Join(ctrPath, "neo-go.yml"))
	e.DeployContract(t, ctrGetNum, nil)
	inv := e.CommitteeInvoker(ctrGetNum.Hash)
	inv.InvokeFail(t, "Cannot get number", "getNumber", validKey)
}

type Coverage struct {
	Document  string
	MethodID  string
	StartLine int
	StartCol  int
	EndLine   int
	EndCol    int
	Counter   int
}

func TestContract_PutNumber(t *testing.T) {
	e := newExecutor(t)
	ctrPutGetNum := neotest.CompileFile(t, e.CommitteeHash, ctrPath, path.Join(ctrPath, "neo-go.yml"))
	e.DeployContract(t, ctrPutGetNum, nil)
	di := ctrPutGetNum.DebugInfo

	// Find contract.go doc
	iDoc := len(di.Documents)
	for i, cDoc := range di.Documents {
		if strings.Contains(cDoc, "contract.go") {
			iDoc = i
		}
	}
	if iDoc == len(di.Documents) {
		fmt.Fprintf(os.Stderr, "No such file\n")
		t.FailNow()
	}

	// Get sequence points for current method
	iPutMethod := len(di.Methods)
	for i := 0; i < len(di.Methods); i++ {
		if di.Methods[i].ID == "PutNumber" {
			iPutMethod = i
			break
		}
	}
	if iPutMethod == len(di.Methods) {
		fmt.Fprintf(os.Stderr, "No \"PutNumber\" method in %s contract\n", ctrPutGetNum.Manifest.Name)
		t.FailNow()
	}
	//seqPoints := make([]compiler.DebugSeqPoint, len(di.Methods[iPutMethod].SeqPoints))
	//copy(seqPoints, di.Methods[iPutMethod].SeqPoints)
	var seqPoints []compiler.DebugSeqPoint
	for _, curPoint := range di.Methods[iPutMethod].SeqPoints {
		if curPoint.Document == iDoc {
			seqPoints = append(seqPoints, curPoint)
		}
	}
	spew.Dump(seqPoints)

	ic, err := e.Chain.GetTestVM(trigger.Application, nil, nil)
	require.NoError(t, err)
	fmt.Println(ic.VM.SyscallHandler == nil)

	ic.VM.LoadNEFMethod(ctrPutGetNum.NEF, ctrPutGetNum.Hash, ctrPutGetNum.Hash, callflag.All, false, 101, -1, nil)
	ic.VM.Context().Estack().PushVal(2)
	ic.VM.Context().Estack().PushVal(validKey)

	fmt.Println("\nPrinting debug info for PutNumber:")
	for !ic.VM.HasStopped() {
		_, curInstr := ic.VM.Context().NextInstr()
		for i := 0; i < len(seqPoints); i++ {
			if int(curInstr) == seqPoints[i].Opcode {
				fmt.Printf("%s: %d.%d, %d.%d hooray!!\n", di.Documents[iDoc], seqPoints[i].StartLine, seqPoints[i].StartCol, seqPoints[i].EndLine, seqPoints[i].EndCol)
			}
		}
		//fmt.Println(nStr, " ", curInstr)
		ic.VM.Step()
	}
	if ic.VM.HasFailed() {
		//govm.SyscallHandler()
		fmt.Fprintln(os.Stderr, "Something went wrong")
		t.FailNow()
	}
	fmt.Println("\n Done printing!")

	// e.DeployContract(t, ctrPutGetNum, nil)
	// // inv := e.CommitteeInvoker(ctrPutGetNum.Hash)
	// // inv.Invoke(t, stackitem.Null{}, "putNumber", validKey, 42)
	// // inv.Invoke(t, 42, "getNumber", validKey)

	// fmt.Println("\n I'm printing debug info for compiled file!")
	// spew.Dump(ctrPutGetNum.DebugInfo)
	// fmt.Println()
	// //fmt.Println(json.MarshalIndent(ctrPutGetNum.Manifest, "", " "))

	// govm := vm.New()
	// govm.LoadNEFMethod(ctrPutGetNum.NEF, ctrPutGetNum.Hash, ctrPutGetNum.Hash, callflag.All, false, 101, -1, nil)
	// fmt.Println("PrintOps:")
	// govm.PrintOps(os.Stdout)
	// govm.Context().Estack().PushVal(2)
	// govm.Context().Estack().PushVal(validKey)

	// fmt.Println("\nPrinting opcodes for PutNumber:")
	// for !govm.HasStopped() {
	// 	nStr, curInstr := govm.Context().NextInstr()
	// 	fmt.Println(nStr, " ", curInstr)
	// 	govm.Step()
	// }
	// if govm.HasFailed() {
	// 	//govm.SyscallHandler()
	// 	fmt.Fprintln(os.Stderr, "Something went wrong")
	// 	t.FailNow()
	// }
	// govm.LoadNEFMethod(ctrPutGetNum.NEF, ctrPutGetNum.Hash, ctrPutGetNum.Hash, callflag.All, true, 0, -1, nil)
	// govm.Context().Estack().PushVal(validKey)
	// fmt.Println("\nPrinting opcodes for GetNumber:")
	// for !govm.HasStopped() {
	// 	nStr, curInstr := govm.Context().NextInstr()
	// 	fmt.Println(nStr, " ", curInstr)
	// 	govm.Step()
	// }
	// res := govm.Estack().Pop().Value()
	// fmt.Println("\nresult = ", res)
	// fmt.Println("\n Done printing!")
}

func TestContract_InvalidKey(t *testing.T) {
	e := newExecutor(t)
	ctrPutGetNum := neotest.CompileFile(t, e.CommitteeHash, ctrPath, path.Join(ctrPath, "neo-go.yml"))
	e.DeployContract(t, ctrPutGetNum, nil)
	inv := e.CommitteeInvoker(ctrPutGetNum.Hash)
	inv.InvokeFail(t, "Invalid key size", "putNumber", invalidKey, 42)
	inv.InvokeFail(t, "Invalid key size", "getNumber", invalidKey)
}

func TestContract_Add(t *testing.T) {
	e := newExecutor(t)
	ctrAdd := neotest.CompileFile(t, e.CommitteeHash, ctrPath, path.Join(ctrPath, "neo-go.yml"))
	e.DeployContract(t, ctrAdd, nil)
	inv := e.CommitteeInvoker(ctrAdd.Hash)
	inv.Invoke(t, 3, "add", 1, 2)

	govm := vm.New()
	govm.LoadNEFMethod(ctrAdd.NEF, ctrAdd.Hash, ctrAdd.Hash, callflag.All, true, 170, -1, nil)
	govm.Context().Estack().PushVal(1)
	govm.Context().Estack().PushVal(2)

	fmt.Println("\nPrinting opcodes:")
	for !govm.HasStopped() {
		nStr, curInstr := govm.Context().NextInstr()
		fmt.Println(nStr, " ", curInstr)
		govm.Step()
	}

	res := govm.Estack().Pop().Value()
	fmt.Println("\nresult = ", res)
}

func Test(t *testing.T) {
	govm := vm.New()
	err := govm.LoadFileWithFlags(path.Join(ctrPath, "contract.nef"), callflag.All)
	if err != nil {
		fmt.Fprintln(os.Stderr, "NEF loading failed")
		t.FailNow()
	}
	govm.PrintOps(os.Stdout)
	govm.Context().Jump(170)
	govm.Context().Estack().PushVal(1)
	govm.Context().Estack().PushVal(2)
	govm.Run()
	res := govm.Estack().Pop().Value()
	fmt.Println(res)
}
